;;;; coerce-support.scm
;;;; Kon Lovett, Sep '06

;;"coerce.scm" Scheme Implementation of COMMON-LISP COERCE and TYPE-OF.
; Copyright (C) 1995, 2001 Aubrey Jaffer
;
;Permission to copy this software, to modify it, to redistribute it,
;to distribute modified versions, and to use it for any purpose is
;granted, subject to the following restrictions and understandings.
;
;1.  Any copy made of this software must include this copyright notice
;in full.
;
;2.  I have made no warranty or representation that the operation of
;this software will be error-free, and I am under no obligation to
;provide any services, by way of maintenance, update, or otherwise.
;
;3.  In conjunction with products arising from the use of this
;material, there shall be no use of my name in any advertising,
;promotional, or sales literature without prior written consent in
;each case.

;; Issues
;;
;; - Cannot know before attempt if coercion possible.

(eval-when (compile)
  (declare
    (usual-integrations)
    (fixnum)
    (inline)
    (no-procedure-checks)
    (no-bound-checks)
    (export
      make-case-coerce
      extend-coerce
      composite-coerce-extension
      coerce-extended?
      coerce-extension
      remove-coerce-extension
      type-of
      coerce
      coerce-all) ) )

(use srfi-1 extras)
(use miscmacros misc-extn-list)

;;;

(define (->boolean x)
  (and x #t) )

#; ; Unused
(define (check-symbol loc obj)
  (unless (symbol? obj)
    (error loc "invalid symbol" obj)) )

(define (check-procedure loc obj)
  (unless (procedure? obj)
    (error loc "invalid procedure" obj)) )

;; Other Type-Of

(define *other-types* '())

(define (set-other-type! typ pred proc)
  (set! *other-types* (alist-update! typ (cons pred proc) *other-types* eq?)) )

(define (unset-other-type! typ)
  (set! *other-types* (alist-delete-first typ *other-types* eq?)) )

(define (other-typdef/test tst)
  (any (lambda (x) (and (tst x) x)) *other-types*) )

(define (other-typdef-info/type typ)
  (alist-ref typ *other-types* eq?) )

;;

(define (typdef-typ td)
  (car td) )

(define (typdef-info td)
  (cdr td) )

(define (typinfo-pred ti)
  (car ti) )

(define (typinfo-proc ti)
  (cdr ti) )

(define (typdef-type? td obj)
  ((typinfo-pred (typdef-info td)) obj) )

;;; Extension

(define (make-case-coerce func . args)
  (let-optionals args ([alist '()])
    (lambda (obj typ err)
      (func obj typ
            (lambda ()
              (if* (alist-ref typ alist eq?) (it obj) (err) ) ) ) ) ) )

(define (coerce-extended? typ)
  #;(check-symbol 'coerce-extended? typ)
  (->boolean (other-typdef-info/type typ)) )

(define (coerce-extension typ)
  #;(check-symbol 'coerce-extension typ)
  (if* (other-typdef-info/type typ)
       (values (typinfo-pred it) (typinfo-proc it))
       (values #f #f) ) )

(define (extend-coerce typ pred proc)
  #;(check-symbol 'extend-coerce typ)
  (check-procedure 'extend-coerce pred)
  (check-procedure 'extend-coerce proc)
  (set-other-type! typ pred proc) )

(define (composite-coerce-extension typ pred proc)
  #;(check-symbol 'extend-coerce typ)
  (check-procedure 'extend-coerce pred)
  (check-procedure 'extend-coerce proc)
  (if* (other-typdef-info/type typ)
       (let ([old-pred (typinfo-pred it)]
             [old-proc (typinfo-proc it)])
         (set-other-type!
           typ
           (if (eq? pred old-pred)
               pred
               (lambda (x) (or (pred x) (old-pred x))) )
           (lambda (obj typ err) (proc obj typ (lambda () (old-proc obj typ err))))) )
       (set-other-type! typ pred proc) ) )

(define (remove-coerce-extension typ)
  (unset-other-type! typ) )

;;; Coerce

;;@body
;;Returns a symbol name for the type of @1.
(define (type-of obj)

  (define (other-type-of obj)
    (and-let* ([td (other-typdef/test (cut typdef-type? <> obj))])
      (typdef-typ td) ) )

  (cond [(boolean? obj)       'boolean]
        [(char? obj)          'char]
        [(number? obj)        'number]
        [(string? obj)        'string]
        [(symbol? obj)        'symbol]
        [(input-port? obj)    'port]
        [(output-port? obj)   'port]
        [(procedure? obj)     'procedure]
        [(eof-object? obj)    'eof-object]
        [(list? obj)          'list]
        [(pair? obj)          'pair]
        [(vector? obj)        'vector]
        [else                 (or (other-type-of obj)
                                  'other)] ) )

;;@body
;;Converts and returns @1 of type @code{char}, @code{number},
;;@code{string}, @code{symbol}, @code{list}, or @code{vector} to
;;@2 (which must be one of these symbols).
(define (coerce obj result-type #!optional default-proc)

  (define (other-coerce objtyp obj restyp err)
    (if* (other-typdef-info/type objtyp)
         ((typinfo-proc it) obj restyp err)
         (err)) )

  (let* (
      [obj-type
        (type-of obj)]
      [err
        (lambda ()
          (if default-proc
              (default-proc obj-type obj result-type)
              (error 'coerce "cannot coerce" obj result-type)))]
      [other->other
        (lambda () (other-coerce obj-type obj result-type err))])
    (if (eq? obj-type result-type)
        obj
        (case obj-type
          [(char)   (case result-type
                      [(number integer) (char->integer obj)]
                      [(string) (string obj)]
                      [(symbol) (string->symbol (string obj))]
                      [(list)   (list obj)]
                      [(vector) (vector obj)]
                      [else     (other->other)])]
          [(number) (case result-type
                      [(char)   (integer->char obj)]
                      [(atom)   obj]
                      [(integer) obj]
                      [(string) (number->string obj)]
                      [(symbol) (string->symbol (number->string obj))]
                      [(list)   (string->list (number->string obj))]
                      [(vector) (list->vector (string->list (number->string obj)))]
                      [else     (other->other)])]
          [(string) (case result-type
                      [(char)   (if (= 1 (string-length obj))
                                    (string-ref obj 0)
                                    (other->other))]
                      [(atom)   (or (string->number obj) (string->symbol obj))]
                      [(number integer) (or (string->number obj) (other->other))]
                      [(symbol) (string->symbol obj)]
                      [(list)   (string->list obj)]
                      [(vector) (list->vector (string->list obj))]
                      [else     (other->other)])]
          [(symbol) (case result-type
                      [(char)   (coerce (symbol->string obj) 'char)]
                      [(number integer) (coerce (symbol->string obj) 'number)]
                      [(string) (symbol->string obj)]
                      [(atom)   obj]
                      [(list)   (string->list (symbol->string obj))]
                      [(vector) (list->vector (string->list (symbol->string obj)))]
                      [else     (other->other)])]
          [(list)   (case result-type
                      [(char)   (if (and (= 1 (length obj))
                                         (char? (car obj)))
                                    (car obj)
                                    (other->other))]
                      [(number integer)
                       (or (string->number (list->string obj)) (other->other))]
                      [(string) (list->string obj)]
                      [(symbol) (string->symbol (list->string obj))]
                      [(vector) (list->vector obj)]
                      [else     (other->other)])]
          [(vector) (case result-type
                      [(char)   (if (and (= 1 (vector-length obj))
                                         (char? (vector-ref obj 0)))
                                    (vector-ref obj 0)
                                    (other->other))]
                      [(number integer)
                       (or (string->number (coerce obj string)) (other->other))]
                      [(string) (list->string (vector->list obj))]
                      [(symbol) (string->symbol (coerce obj string))]
                      [(list)   (list->vector obj)]
                      [else     (other->other)])]
          [else (other->other)] ) ) ) )

;;

(define (coerce-all flst tlst #!optional default-proc)
  (map (cut coerce <> <> default-proc) flst tlst) )
